/*********************************************************************
 * TTCE 0.5.0: Taxes, Tips and Currencies in Euro
 * (C) Jrme Euzenat, 1998, 2000-2002
 * $Id: ttce.c,v 1.19 2002/08/19 21:26:32 euzenat Exp $
 *
 * Parts:
 * (C) Alan Harder, 1999
 * From TipMe v1.23 - 3/29/99
 * Alan.Harder@Sun.COM     http://www.moshpit.org/pilot/
 * Freeware - feel free to distribute and/or modify this code.. just
 *            leave this header on it.
 ********************************************************************/

/************************
 * CODING ISSUES
 * We want to optimize size, not speed
 *
 * Ensure that for each New, there is a Free.
 */

#include "ttce.h"
#include <GuiTools.h>
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#define TRUE 1
#define FALSE 0
#define ERROR -1
#define MAIN 0

/***********************************************************************
 * Application
 ***********************************************************************/

#define APP_ID		'TTCE'

#define CREATOR		'TTCE'

/***********************************************************************
 * database structures
 ***********************************************************************/

#define CARD_NO		0
#define DB_NAME		"TaxTipsCurrencies"
#define DB_TYPE		'Data'
#define DB_VERSION	2
/* 1= initial
   2= value as Long
*/

UInt recIndex;
VoidHand recordH;
DmOpenRef db;

#define MAXITEMS 128

#define TAXCAT 0
#define TIPCAT 1
#define CURRCAT 2

UInt currentCategory = TAXCAT;

#define TAXCAT_NAME "tax"
#define TIPCAT_NAME "tip"
#define CURRCAT_NAME "currency"

/* Header pour les categories */
AppInfoType appinfo = {
  0,  /* Pas compris... */
  /* Liste de toutes les categories */
  TAXCAT_NAME, TIPCAT_NAME, CURRCAT_NAME, "", "", "",  "", "", "", "", "", "", "", "", "", "",
  /* Category Uniq IDs */
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
  /* Last Uniq ID */
  15
};

typedef struct
	{
	Long		value;		// not int
	UInt		opt;		// for the tip and tax
	Char		name[MAXNAME];  // missnamed
	} Item;
// opt: if tax=TRUE/FALSE depending on "apply to total"
// 	if tip=selections of the chooser:
//	food = (item.opt & foodf) << 6
//	place = (item.opt & placef) << 3
//	mood = (item.opt & moodf)

Item *currentItem;
	
// values are coded over 3 bits (0 1 2 3 4)
#define fullMask 7
#define fullSet fullMask+(fullMask<<3)+(fullMask<<6)

#define food(x) (x->opt & fullMask)
#define serv(x) ((x->opt >> 3) & fullMask)
#define place(x) ((x->opt >> 6) & fullMask)

/***********************************************************************
 * Preferences
 ***********************************************************************/

#define PREF_VERSION	2
#define PREF_ID		0

// App Preferences Type
typedef struct
{
    Int taxRate, tipRate, tax2Rate, currRate;
    Char taxIt[MAXNAME], tipIt[MAXNAME], tax2It[MAXNAME], currIt[MAXNAME], curr2It[MAXNAME];
    Boolean useTax;
}
TtcePrefsType;
// Old preferences for installing V2 (i.e. >= 0.4.0) over an older version
typedef struct
{
    Int taxRate, tipRate, tax2Rate, currRate,  taxIt, tipIt, tax2It, currIt, curr2It;
    Boolean useTax;
}
TtcePrefsV1Type;

TtcePrefsType *prefs, *storedPrefs;

void getCurrentPrefs( TtcePrefsType *pr, UInt16 prSize, Boolean stored){
    TtcePrefsType *oldprefs;
    UInt prefversion = PrefGetAppPreferences(CREATOR, PREF_ID, pr, &prSize, stored);
    if ( prefversion != PREF_VERSION ){
    	if ( prefversion == noPreferenceFound){
	  pr->tipRate = pr->taxRate = pr->tax2Rate = 0;
	  pr->currRate = 100;
	  pr->useTax = 0;
	} else {
	  // Old preference version
	  LocalID db_id = DmFindDatabase(CARD_NO, DB_NAME);
	  // If exists a v1-database, destroy it
	  if (db_id) DmDeleteDatabase( CARD_NO, db_id );
          // Like a bug
	  //oldprefs = pr;
	  //pr = MemPtrNew(prSize);
	  //pr->tipRate = oldprefs->tipRate;
	  //pr->taxRate = oldprefs->taxRate;
	  //pr->tax2Rate = oldprefs->tax2Rate;
	  //pr->currRate = oldprefs->currRate;
	  //pr->useTax = oldprefs->useTax;
	  //MemPtrFree(oldprefs);
	}
        StrCopy( pr->tipIt, "none" );
        StrCopy( pr->taxIt, "none" );
        StrCopy( pr->tax2It, "none" );
	StrCopy( pr->currIt, "EUR" );
	StrCopy( pr->curr2It, "EUR" );
    }
}

/***********************************************************************
 * database management
 ***********************************************************************/

Boolean newP = FALSE;

UInt newRecord( CharPtr name, Long value, UInt option, UInt category) {
  UInt record_attr ;
  Item ss, *rec;
  ss.value = value;
  ss.opt = option;
  StrCopy(ss.name,name);
  recIndex = dmMaxRecordIndex; //at the end of the DB (faster)
  recordH = DmNewRecord( db, &recIndex, sizeof(Item)+1);
  DmWrite( MemHandleLock(recordH), 0, &ss, sizeof(Item) );
  MemHandleUnlock(recordH);
  DmRecordInfo (db, recIndex, &record_attr, NULL, NULL) ;
  record_attr = ( record_attr & ~dmRecAttrCategoryMask) | category;
  DmSetRecordInfo (db, recIndex, &record_attr, NULL) ;
  DmReleaseRecord( db, recIndex, true );
  return recIndex;
}

void initDB() {
  /*
  // This could be replaced by the use of a dbimage
  // but it is bigger?
  VoidHand imageH = DmGetResource('DBIM',DBImage);
  DmCreateDatabaseFromImage(MemHandleLock(imageH));
  MemHandleUnlock(imageH);
  DmReleaseResource(imageH);
  db = DmOpenDatabaseByTypeCreator( DBTYPE, CREATOR, dmModeReadWrite);
  */
  newRecord( "none", 0, 0, TAXCAT);
  newRecord( "TVA", 1960, 0, TAXCAT);
  newRecord( "TVA5", 500, 1, TAXCAT);
  newRecord( "none", 0, 0, TIPCAT);
  newRecord( "Serv", 1500, 2+(2<<3)+(2<<6), TIPCAT);
  newRecord( "EUR", 100, 0, CURRCAT);
}

void openDB()
 {
  LocalID db_id ;
  UInt db_attrs ;
  LocalID appInfoID ;
  
  // In case DatabaseInfo is not needed, it is possible to drop db_id
  db_id = DmFindDatabase (CARD_NO, DB_NAME) ;
  if (!db_id) {
    DmCreateDatabase (CARD_NO, DB_NAME, APP_ID, DB_TYPE, false) ;
    db_id = DmFindDatabase (CARD_NO, DB_NAME);
    db = DmOpenDatabase (CARD_NO, db_id, dmModeReadWrite);
    initDB(db_id);
  } else {
    db = DmOpenDatabase(CARD_NO, db_id, dmModeReadWrite) ;
  };
  if (!db) return;

  DmDatabaseInfo (CARD_NO, db_id, NULL, &db_attrs, NULL, NULL, NULL, 
		  NULL, NULL, &appInfoID, NULL, NULL, NULL) ;
  if (appInfoID == NULL) {
    VoidHand h = DmNewHandle(db, sizeof(AppInfoType)) ;
    AppInfoType *appInfoP;
    appInfoID = MemHandleToLocalID(h) ;
    DmSetDatabaseInfo (CARD_NO, db_id, NULL, NULL, NULL, NULL, NULL, NULL,
		       NULL, &appInfoID, NULL, NULL, NULL) ;
    appInfoP = MemLocalIDToLockedPtr(appInfoID, 0) ;
    // bizarre: the initial appinfo
    DmWrite(appInfoP, 0, &appinfo, sizeof(AppInfoType)) ;
    MemPtrUnlock(appInfoP) ;
  }

  db_attrs = db_attrs | dmHdrAttrBackup ; //mark as to be sync'ed
  DmSetDatabaseInfo (CARD_NO, db_id, NULL, &db_attrs, NULL, NULL, NULL,
		     NULL, NULL, NULL, NULL, NULL, NULL) ;
}

VoidHand findRecordByRank( UInt rank, UInt category ) {
     UInt i, j=0;

     for( i=0; i<rank; i++, j++) DmQueryNextInCategory( db, &j, category );
     return ( DmQueryNextInCategory( db, &j, category ) );
}

// Returns j instead of Handler
UInt findRecIndexByRank( UInt rank, UInt category ) {
     UInt i, j=0;

     for( i=0; i<rank+1; i++, j++) DmQueryNextInCategory( db, &j, category );
     return ( --j );
}

UInt findRecRankByName( Char *name, UInt category ) {
     UInt i, j=0;
     UInt nbItems = DmNumRecordsInCategory(db, category);
     Boolean found = false;

     for( i=0; i<nbItems && !found; i++, j++) {
	recordH = DmQueryNextInCategory( db, &j, category );
	currentItem = MemHandleLock(recordH);
	if ( StrCompare( currentItem->name, name) == 0) found = true;
	MemHandleUnlock(recordH);
     }
     return ( found?--i:0 );
}

/***********************************************************************
 * database and interface
 ***********************************************************************/

/****************** Amount field management *****************/

Char deciSep[] = ".", deciNoDime[] = ".0";
Char deciChr = '.';

// Methods to get and set Amounts in text fields.  An amount is the
// number of CENTS, and will be displayed with the decimal point placed
// appropriately..

void PutAmountInField(FieldPtr fieldPtr, Long amount)
{
    CharPtr str;
    Long amt;

    str = (CharPtr)MemPtrNew(16);
    if (amount < 0)
    {
	amt = -amount;
	StrCopy(str, "(");
    }
    else
    {
	amt = amount;
	StrCopy(str, "");
    }
    StrIToA(str + StrLen(str), amt/100);
    StrCat(str, (amt%100 < 10) ? deciNoDime : deciSep);
    StrIToA(str + StrLen(str), amt%100);
    if (amount < 0) StrCat(str, ")");
    SetText(fieldPtr, str);
    MemPtrFree(str);
}

Long GetAmountFromField(FieldPtr fieldPtr)
{
    Long val;
    Int cents;
    CharPtr str, str2;
    str = GetText(fieldPtr);
    if (str == NULL) return ERROR;
    if (StrLen(str) == 0) { MemPtrFree(str); return ERROR; }
    val = StrAToI(str) * 100;
    str2 = StrChr(str, deciChr);
    if (str2 != NULL && str2[1] >= '0' && str2[1] <= '9')
    {
	cents = (str2[1] - '0') * 10;
	if (str2[2] >= '0' && str2[2] <= '9')
	    cents += str2[2] - '0';
	val += cents;
    }
    MemPtrFree(str);
    return val;
}

/****************** Dynamic chooser management *****************/

#define MAXVISIBLEITEMS 8

UInt initListChoices( UInt category, CharPtr list ){
  UInt nbItems = DmNumRecordsInCategory(db, category);
  UInt i;
  CharPtr *listitems = (CharPtr *)MemPtrNew((nbItems+1)*sizeof(CharPtr));
  
  recIndex = 0;
  for( i=0; i<nbItems; i++) {
    recordH = DmQueryNextInCategory( db, &recIndex, category );
    currentItem = MemHandleLock(recordH);
    listitems[i] = (CharPtr)MemPtrNew((StrLen(currentItem->name)+1)*sizeof(char));
    StrCopy(listitems[i],currentItem->name);
    MemHandleUnlock(recordH);
    recIndex++;
    }
  listitems[nbItems] = 0;

//  list = &listitems;
  return nbItems;
}

void initChooser( ListPtr list, UInt category, Boolean firstTime ) {
  UInt nbItems = DmNumRecordsInCategory(db, category);
  UInt i;
  CharPtr *listitems = (CharPtr *)MemPtrNew((nbItems+1)*sizeof(CharPtr));

  recIndex = 0;
  for( i=0; i<nbItems; i++) {
    recordH = DmQueryNextInCategory( db, &recIndex, category );
    currentItem = MemHandleLock(recordH);
    listitems[i] = (CharPtr)MemPtrNew((StrLen(currentItem->name)+1)*sizeof(char));
    StrCopy(listitems[i],currentItem->name);
    MemHandleUnlock(recordH);
    recIndex++;
    }
  listitems[nbItems] = 0;

  LstSetListChoices( list, listitems, nbItems );
  // It seems from the doc (p329) that POS takes care of freeing the old list

  if( list->id != itemsAvailable ){
    if( nbItems < MAXVISIBLEITEMS ) LstSetHeight( list, nbItems );
    else LstSetHeight( list, MAXVISIBLEITEMS );
  }
  if ( firstTime ) LstSetSelection( list, 0 );
  else LstDrawList( list );
}

void setTriggeredList( ListPtr list, ControlPtr trig, UInt rank ){
  LstSetSelection( list, rank );
  CtlSetLabel( trig, LstGetSelectionText( list, rank ) );
}

void updateChooser( ListPtr list, ControlPtr trig, FieldPtr fld, ControlPtr cbox, UInt category, int nbItems, CharPtr *listItems) {
  Char buf[MAXNAME];
  UInt rank;
 
  StrCopy( buf, LstGetSelectionText(list, LstGetSelection(list)) );
  LstSetListChoices( list, listItems, nbItems );
  // It seems from the doc (p329) that POS takes care of freeing the old list
  if( nbItems < MAXVISIBLEITEMS ) LstSetHeight( list, nbItems );
  else LstSetHeight( list, MAXVISIBLEITEMS );
  rank = findRecRankByName(buf,category);
  setTriggeredList( list, trig, rank);
  if ( fld != NULL && rank != 0 ){ // Do not change value if selection not found
    recordH = findRecordByRank( rank, category );
    currentItem = MemHandleLock(recordH);
    PutAmountInField( fld, currentItem->value );
    MemHandleUnlock(recordH);
  }
  if ( cbox != NULL && rank != 0 ){ // Do not change value if selection not found
    recordH = findRecordByRank( rank, category );
    currentItem = MemHandleLock(recordH);
    CtlSetValue( cbox, currentItem->opt );
    MemHandleUnlock(recordH);
  }
}

/***********************************************************************
 * interface structures
 ***********************************************************************/

Long total, amount, tip, tax, tax2, tipbase, ctotal;
FieldPtr amountFld, taxFld, taxAmtFld, tax2Fld, tax2AmtFld,
         tipFld, tipAmtFld, totalFld, currFld, currAmtFld;
ControlPtr taxOnAll;
Word currentField = NULL, lastField=NULL;

static CharPtr currSymbols[5] = {"\200", "$", "", "", ""};
static int currIndex = 0;

/***********************************************************************
 * event handlers for DB edition
 ***********************************************************************/

UInt currentEditor;

Boolean toupdate = false;

void fillEditor(UInt editorID){
  FormPtr f = FrmGetActiveForm();

  SetText((FieldPtr)FrmGetObjectPtrById(f, nameEditField), currentItem->name);
  PutAmountInField((FieldPtr)FrmGetObjectPtrById(f, valueEditField),currentItem->value);
  switch (editorID)
  {
    case taxEditForm: 
	CtlSetValue( (ControlPtr)FrmGetObjectPtrById(f, taxCBox), currentItem->opt );
	break;
    case tipEditForm:
	setTriggeredList( (ListPtr)FrmGetObjectPtrById(f, foodTipList), 
			  (ControlPtr)FrmGetObjectPtrById(f, foodTipTrigger),
			   food(currentItem));
	setTriggeredList( (ListPtr)FrmGetObjectPtrById(f, servTipList), 
			  (ControlPtr)FrmGetObjectPtrById(f, servTipTrigger),
			   serv(currentItem));
	setTriggeredList( (ListPtr)FrmGetObjectPtrById(f, placeTipList), 
			  (ControlPtr)FrmGetObjectPtrById(f, placeTipTrigger),
			   place(currentItem));
	break;
    case currEditForm:
	PutAmountInField((FieldPtr)FrmGetObjectPtrById(f, currRevField),(10000/currentItem->value));
	break;
  }
}

void fillDBWithValue(UInt editorID){
  FormPtr f = FrmGetActiveForm();
  UInt record_attr ;

  if( !newP ) currentItem = MemPtrNew(sizeof(Item));
  StrCopy( currentItem->name, GetText((FieldPtr)FrmGetObjectPtrById(f, nameEditField)));
  currentItem->value = GetAmountFromField((FieldPtr)FrmGetObjectPtrById(f, valueEditField));
  switch (editorID)
  {
    case taxEditForm: 
	currentItem->opt = CtlGetValue( (ControlPtr)FrmGetObjectPtrById(f, taxCBox));
	break;
    case tipEditForm: 
	currentItem->opt = (LstGetSelection((ListPtr)FrmGetObjectPtrById(f, placeTipList))<<6)
                         | (LstGetSelection((ListPtr)FrmGetObjectPtrById(f, servTipList))<<3)
                         | LstGetSelection((ListPtr)FrmGetObjectPtrById(f, foodTipList));
	break;
    case currEditForm:
	//BUG: I must absolutely test that value is not 0!
	break;
  }

  if( newP ){ // It is time to declare this to the DB
    recIndex = dmMaxRecordIndex; //at the end of the DB (faster)
    recordH = DmNewRecord( db, &recIndex, sizeof(Item)+1);
    DmRecordInfo (db, recIndex, &record_attr, NULL, NULL) ;
    record_attr = ( record_attr & ~dmRecAttrCategoryMask) | currentCategory;
    DmSetRecordInfo (db, recIndex, &record_attr, NULL) ;
  } else { // It is time to get the record
    MemHandleUnlock(recordH);
    recordH = DmGetRecord(db, recIndex );
  }
    DmWrite( MemHandleLock(recordH), 0, currentItem, sizeof(Item) );
    MemHandleUnlock(recordH);
    DmReleaseRecord( db, recIndex, true );
    MemPtrFree(currentItem);
}

// Item editors
Boolean EditFormEventHandler( EventType *event ){ 
  FormPtr f = FrmGetActiveForm();
  switch (event->eType)
  {
    case frmLoadEvent: 
      break;
    case frmOpenEvent:
      currentEditor = event->data.frmOpen.formID;
      if( newP ) {
	currentItem = MemPtrNew(sizeof(Item));
	StrCopy(currentItem->name, "");
	if ( currentEditor != currEditForm ) currentItem->value = 0;
	else currentItem->value = 100;
	currentItem->opt = 0;
      } else currentItem = MemHandleLock( recordH );
      fillEditor(currentEditor);
      FrmDrawForm(f);
      if ( newP ) FrmSetFocusById(f, nameEditField);
      else FrmSetFocusById(f, valueEditField); 
    return TRUE;

  case ctlSelectEvent:
    switch (event->data.ctlSelect.controlID)
    {
    case cancelEditButton:
      if( newP ) MemPtrFree( currentItem );
      else MemHandleUnlock(recordH);
      newP = FALSE;
      FrmReturnToForm(editListForm);
      return TRUE;
    case resetEditButton:
      fillEditor(currentEditor);
      FrmDrawForm(f);
      return TRUE;
    case saveEditButton:
      if( !newP ){
	MemHandleUnlock(recordH);
      }
      fillDBWithValue(currentEditor);
      initChooser( (ListPtr)FrmGetObjectPtrById( FrmGetFormPtr(editListForm), itemsAvailable), currentCategory, false );
      toupdate = true;
      newP = FALSE;
      FrmUpdateForm( editListForm, frmRedrawUpdateCode );
      FrmReturnToForm(editListForm);
      return TRUE;

    // There are also fields, checkBox and popUpLists...
    //-- For the taxes we must have
    case taxCBox:
	// nothing to be done
      return TRUE;
    //-- For the currency we must have
    case pivotCurrSymbolField:
      currIndex = (currIndex+1) % 5;
      CtlSetLabel( (ControlPtr)FrmGetObjectPtrById(FrmGetActiveForm(), pivotCurrSymbolField), currSymbols[currIndex]);
      return TRUE;
    //-- For the tip  we must have
    case placeTipTrigger:
    case servTipTrigger:
    case foodTipTrigger:
	FrmHandleEvent(FrmGetActiveForm(), event); // default action
	// The BigTipper method is piecewise manhattan
	//0-3-5-6-8
	// The global method (JE pattent)
	PutAmountInField( (FieldPtr)FrmGetObjectPtrById(f, valueEditField),
		(LstGetSelection((ListPtr)FrmGetObjectPtrById(f, foodTipList))
		+LstGetSelection((ListPtr)FrmGetObjectPtrById(f, servTipList))
		+LstGetSelection((ListPtr)FrmGetObjectPtrById(f, placeTipList)))
		* 250 ); // hope this can be a Long?
	FldDrawField((FieldPtr)FrmGetObjectPtrById(f, valueEditField));
      return TRUE;
    }
    // deal with the RevEdit and ... in currency

    case penDownEvent:
	lastField = currentField;
	currentField = NULL;
	return FALSE;

    case fldEnterEvent:
	currentField = event->data.fldEnter.fieldID;
	FrmHandleEvent(FrmGetActiveForm(), event); // default action
        return TRUE;

    case penUpEvent:
	if (lastField)
	  if ( lastField == valueEditField ){
	    PutAmountInField((FieldPtr)FrmGetObjectPtrById(f, currRevField),(10000/GetAmountFromField((FieldPtr)FrmGetObjectPtrById(f, valueEditField))));
	   FldDrawField((FieldPtr)FrmGetObjectPtrById(f, currRevField));
        } else {
	    PutAmountInField((FieldPtr)FrmGetObjectPtrById(f, valueEditField),(10000/GetAmountFromField((FieldPtr)FrmGetObjectPtrById(f, currRevField))));
	   FldDrawField((FieldPtr)FrmGetObjectPtrById(f, valueEditField));
        }
	FrmHandleEvent(FrmGetActiveForm(), event); // default action
        return TRUE;
  }
  return FALSE;
}



/***********************************************************************
 * computations
 ***********************************************************************/

// JE: This is not correct with regard to the tax2CBox
#define RETROCOMPALPHA() amount=(10000*total)/(10000+GetAmountFromField(taxFld)+GetAmountFromField(tax2Fld)+GetAmountFromField(tipFld));

void computeValues() {
	Long taxRate, tax2Rate, tipRate, currRate;
    amount = GetAmountFromField(amountFld);
    if ( (taxRate = GetAmountFromField(taxFld)) > 0) tax = (amount*taxRate+5000)/10000;	// standard rounding
    else  tax = 0;
    PutAmountInField(taxAmtFld, tax); //FldDrawField(taxAmtFld);
    if ( (tax2Rate = GetAmountFromField(tax2Fld)) > 0)
	if ( CtlGetValue( taxOnAll ) )
		tax2 = ((amount+tax)*tax2Rate+5000)/10000;// standard rounding
	else tax2 = (amount*tax2Rate+5000)/10000;	// standard rounding
    else  tax2 = 0;
    PutAmountInField(tax2AmtFld, tax2);
    if ( (tipRate = GetAmountFromField(tipFld)) > 0) tip = (amount*tipRate+9999)/10000;	// always round up
    else  tip = 0;
    PutAmountInField(tipAmtFld, tip);
    total = amount + tax + tax2 + tip;
    PutAmountInField(totalFld, total);
    if ( (currRate = GetAmountFromField(currFld)) > 0) ctotal = total * currRate/100;
    PutAmountInField(currAmtFld, ctotal);
    FrmUpdateForm( mainForm, frmRedrawUpdateCode );
}

void computeCurrRate(){
  FormPtr frm = FrmGetFormPtr(mainForm);
  ListPtr currLst = FrmGetObjectPtrById( frm, currList );
  ListPtr curr2Lst = FrmGetObjectPtrById( frm, curr2List );
  Long v;

  recordH = findRecordByRank( LstGetSelection(currLst), CURRCAT );
  currentItem = MemHandleLock(recordH);
  v = currentItem->value;
  MemHandleUnlock(recordH);
  recordH = findRecordByRank( LstGetSelection(curr2Lst), CURRCAT );
  currentItem = MemHandleLock(recordH);
  PutAmountInField( currFld, (v*100)/(currentItem->value) );
  MemHandleUnlock(recordH);
}

/***********************************************************************
 * main interface function
 ***********************************************************************/

Boolean MenuEventHandler(EventPtr event)
{
    switch ( event->data.menu.itemID ) {
      case menuItemAbout:
	FrmPopupForm(aboutForm) ;
        return TRUE;
      //case menuItemSplit:
      case menuItemEdit:
	toupdate = false;
	FrmPopupForm(editListForm) ;
        return TRUE;
      case menuItemCut:
      case menuItemCopy:
      case menuItemPaste:
	{
	    FormPtr form ;
	    Word tmp_index ;
	    FieldPtr tmp_fld ;
	    
	    /* JE//: from FP, certainly size optimization that I do not know */
	    form = FrmGetActiveForm () ;
	    tmp_index = FrmGetFocus (form) ;
	    if (tmp_index != -1) {
		/* On croise les doigts pour qu'il n'y ait que les */
		/* fields qui puissent avoir le focus !!! */
		tmp_fld = FrmGetObjectPtr (form, tmp_index) ;
		switch (event->data.menu.itemID) {
		  case menuItemCut: FldCut (tmp_fld) ; break ;
		  case menuItemPaste: FldPaste (tmp_fld) ; break ;
		  case menuItemCopy: FldCopy (tmp_fld) ; break ;
		}
	    }
	    return TRUE;
	}
      default:
	FrmHandleEvent(FrmGetActiveForm(), event); /// This should be useless! Test?
        return TRUE;
        //return FALSE;
    };
    return FALSE;
}

void clearMainForm( FormPtr f ){
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, currList),
		    (ControlPtr)FrmGetObjectPtrById(f, currTrigger),
		    0);
  PutAmountInField(amountFld, 0);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, taxList),
		    (ControlPtr)FrmGetObjectPtrById(f, taxTrigger),
		    0);
  PutAmountInField(taxFld, 0);
  PutAmountInField(taxAmtFld, 0);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, tax2List),
		    (ControlPtr)FrmGetObjectPtrById(f, tax2Trigger),
		    0);
  PutAmountInField(tax2Fld, 0);
  PutAmountInField(tax2AmtFld, 0);
  CtlSetValue( taxOnAll, 0 );
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, tipList),
		    (ControlPtr)FrmGetObjectPtrById(f, tipTrigger),
		    0);
  PutAmountInField(tipFld, 0);
  PutAmountInField(tipAmtFld, 0);
  PutAmountInField(totalFld, 0);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, curr2List),
		    (ControlPtr)FrmGetObjectPtrById(f, curr2Trigger),
		    0);
  PutAmountInField(currFld, 100);
  PutAmountInField(currAmtFld, 0);
}

void storePrefs(TtcePrefsType *pr){
	FormPtr frm = FrmGetActiveForm();

	StrCopy( pr->currIt, LstGetSelectionText((ListPtr)FrmGetObjectPtrById(frm, currList), LstGetSelection((ListPtr)FrmGetObjectPtrById(frm, currList))) );
	pr->taxRate = GetAmountFromField(taxFld);
	StrCopy( pr->taxIt, LstGetSelectionText((ListPtr)FrmGetObjectPtrById(frm, taxList), LstGetSelection((ListPtr)FrmGetObjectPtrById(frm, taxList))) );
	pr->tax2Rate = GetAmountFromField(tax2Fld);
	StrCopy( pr->tax2It, LstGetSelectionText((ListPtr)FrmGetObjectPtrById(frm, tax2List), LstGetSelection((ListPtr)FrmGetObjectPtrById(frm, tax2List))) );
	pr->useTax = CtlGetValue( (ControlPtr)FrmGetObjectPtrById(frm, tax2CBox) );
	pr->tipRate = GetAmountFromField(tipFld);
	StrCopy( pr->tipIt, LstGetSelectionText((ListPtr)FrmGetObjectPtrById(frm, tipList), LstGetSelection((ListPtr)FrmGetObjectPtrById(frm, tipList))) );
	pr->currRate = GetAmountFromField(currFld);
	StrCopy( pr->curr2It, LstGetSelectionText((ListPtr)FrmGetObjectPtrById(frm, curr2List), LstGetSelection((ListPtr)FrmGetObjectPtrById(frm, curr2List))) );
}

void restorePrefs(TtcePrefsType *pr){
  FormPtr f = FrmGetActiveForm();

  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, currList),
		    (ControlPtr)FrmGetObjectPtrById(f, currTrigger),
		    findRecRankByName(pr->currIt, CURRCAT));
  if (pr->taxRate > 0) PutAmountInField(taxFld, pr->taxRate);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, taxList),
		    (ControlPtr)FrmGetObjectPtrById(f, taxTrigger),
		    findRecRankByName(pr->taxIt, TAXCAT));
  if (pr->tax2Rate > 0) PutAmountInField(tax2Fld, pr->tax2Rate);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, tax2List),
		    (ControlPtr)FrmGetObjectPtrById(f, tax2Trigger),
		    findRecRankByName(pr->tax2It, TAXCAT));
  CtlSetValue( (ControlPtr)FrmGetObjectPtrById(f, tax2CBox), pr->useTax );
  if (pr->tipRate > 0) PutAmountInField(tipFld, pr->tipRate);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, tipList),
		    (ControlPtr)FrmGetObjectPtrById(f, tipTrigger),
		    findRecRankByName(pr->tipIt, TIPCAT));
  if (pr->currRate > 0) PutAmountInField(currFld, pr->currRate);
  setTriggeredList( (ListPtr)FrmGetObjectPtrById( f, curr2List),
		    (ControlPtr)FrmGetObjectPtrById(f, curr2Trigger),
		    findRecRankByName(pr->curr2It, CURRCAT));
}

// Reset the triggers from the database
void restoreMainForm(FormPtr frm){
  UInt nbCurr, nbTax, nbTip;
  CharPtr *currItemList, *taxItemList, *tipItemList;

  UInt nbItems , i;
  CharPtr *listitems;

  nbItems = DmNumRecordsInCategory(db, CURRCAT);
  listitems = (CharPtr *)MemPtrNew((nbItems+1)*sizeof(CharPtr));
  
  recIndex = 0;
  for( i=0; i<nbItems; i++) {
    recordH = DmQueryNextInCategory( db, &recIndex, CURRCAT );
    currentItem = MemHandleLock(recordH);
    listitems[i] = (CharPtr)MemPtrNew((StrLen(currentItem->name)+1)*sizeof(char));
    StrCopy(listitems[i],currentItem->name);
    MemHandleUnlock(recordH);
    recIndex++;
    }
  listitems[nbItems] = 0;

  currItemList = listitems;
  nbCurr = nbItems;

  //nbCurr = initListChoices( CURRCAT, &currItemList );
  nbItems = DmNumRecordsInCategory(db, TAXCAT);
  listitems = (CharPtr *)MemPtrNew((nbItems+1)*sizeof(CharPtr));
  
  recIndex = 0;
  for( i=0; i<nbItems; i++) {
    recordH = DmQueryNextInCategory( db, &recIndex, TAXCAT );
    currentItem = MemHandleLock(recordH);
    listitems[i] = (CharPtr)MemPtrNew((StrLen(currentItem->name)+1)*sizeof(char));
    StrCopy(listitems[i],currentItem->name);
    MemHandleUnlock(recordH);
    recIndex++;
    }
  listitems[nbItems] = 0;

  taxItemList = listitems;
  nbTax = nbItems;

  //nbTax = initListChoices( TAXCAT, &taxItemList );
  nbItems = DmNumRecordsInCategory(db, TIPCAT);
  listitems = (CharPtr *)MemPtrNew((nbItems+1)*sizeof(CharPtr));
  
  recIndex = 0;
  for( i=0; i<nbItems; i++) {
    recordH = DmQueryNextInCategory( db, &recIndex, TIPCAT );
    currentItem = MemHandleLock(recordH);
    listitems[i] = (CharPtr)MemPtrNew((StrLen(currentItem->name)+1)*sizeof(char));
    StrCopy(listitems[i],currentItem->name);
    MemHandleUnlock(recordH);
    recIndex++;
    }
  listitems[nbItems] = 0;

  tipItemList = listitems;
  nbTip = nbItems;

  //nbTip = initListChoices( TIPCAT, &tipItemList );

  updateChooser((ListPtr)FrmGetObjectPtrById(frm, currList),
                   (ControlPtr)FrmGetObjectPtrById(frm, currTrigger),
                   (FieldPtr)NULL, (ControlPtr)NULL,
 		   CURRCAT, nbCurr, currItemList );
  updateChooser((ListPtr)FrmGetObjectPtrById(frm, taxList),
                   (ControlPtr)FrmGetObjectPtrById(frm, taxTrigger),
                   taxFld, (ControlPtr)NULL,
 		   TAXCAT, nbTax, taxItemList );
  updateChooser((ListPtr)FrmGetObjectPtrById(frm, tax2List),
                   (ControlPtr)FrmGetObjectPtrById(frm, tax2Trigger),
                   tax2Fld, taxOnAll,
 		   TAXCAT, nbTax, taxItemList );
  updateChooser((ListPtr)FrmGetObjectPtrById(frm, tipList),
                   (ControlPtr)FrmGetObjectPtrById(frm, tipTrigger),
                   tipFld, (ControlPtr)NULL,
 		   TIPCAT, nbTip, tipItemList );
  updateChooser((ListPtr)FrmGetObjectPtrById(frm, curr2List),
                   (ControlPtr)FrmGetObjectPtrById(frm, curr2Trigger),
                   (FieldPtr)NULL, (ControlPtr)NULL,
 		   CURRCAT, nbCurr, currItemList );
  computeCurrRate();
  computeValues();
}

void updateMainPanel( Word fieldID )
{ FormPtr frm = FrmGetActiveForm();
	    switch ( fieldID ) {
	      case taxAmountField:
		tax = GetAmountFromField(taxAmtFld);
		if(amount>0) {
		    PutAmountInField(taxFld, tax*10000/amount);
		    setTriggeredList( (ListPtr)FrmGetObjectPtrById( frm, taxList),
				(ControlPtr)FrmGetObjectPtrById(frm, taxTrigger),
				0);
		}
		break;
	      case tax2AmountField:
		tax2 = GetAmountFromField(tax2AmtFld);
		//x/ if((amount+tax)>0)PutAmountInField(tax2Fld,tax2*10000/(amount+tax));
		if(amount>0) {
		    PutAmountInField(tax2Fld,tax2*10000/amount);
		    setTriggeredList( (ListPtr)FrmGetObjectPtrById( frm, tax2List),
				(ControlPtr)FrmGetObjectPtrById(frm, tax2Trigger),
				0);
		}
		break;
	      case tipAmountField:
		tip = GetAmountFromField(tipAmtFld);
		if(amount>0) {
		    PutAmountInField(tipFld,tip*10000/amount);
		    setTriggeredList( (ListPtr)FrmGetObjectPtrById( frm, tipList),
				(ControlPtr)FrmGetObjectPtrById(frm, tipTrigger),
				0);
		}
		break;
	      case totalField:
		total = GetAmountFromField(totalFld);
		RETROCOMPALPHA()
		PutAmountInField(amountFld,amount);
		break;
	      case curr2AmountField:
		ctotal = GetAmountFromField(currAmtFld);
		if(GetAmountFromField(currFld)>0)total=ctotal/GetAmountFromField(currFld)*100;
		PutAmountInField(totalFld,total);
		RETROCOMPALPHA()
		PutAmountInField(amountFld,amount);
		break;
	    }
	    computeValues();
}

Boolean MainFormEventHandler(EventPtr event)
{ FormPtr frm;

	switch (event->eType)
	{
	  case frmLoadEvent:
	    frm = FrmGetActiveForm();
	    if (event->data.frmLoad.formID == mainForm) {
		//CH: 
		if ( amountFld == NULL ){
		    //CH: Might be time for the first init
		    amountFld = (FieldPtr)FrmGetObjectPtrById(frm, amountField);
		    taxFld = (FieldPtr)FrmGetObjectPtrById(frm, taxRateField);
		    taxAmtFld = (FieldPtr)FrmGetObjectPtrById(frm,taxAmountField);
		    tax2Fld = (FieldPtr)FrmGetObjectPtrById(frm, tax2RateField);
		    tax2AmtFld = (FieldPtr)FrmGetObjectPtrById(frm,tax2AmountField);
		    taxOnAll = (ControlPtr)FrmGetObjectPtrById(frm,tax2CBox);
		    tipFld = (FieldPtr)FrmGetObjectPtrById(frm, tipRateField);
		    tipAmtFld = (FieldPtr)FrmGetObjectPtrById(frm,tipAmountField);
		    totalFld = (FieldPtr)FrmGetObjectPtrById(frm, totalField);
		    currFld = (FieldPtr)FrmGetObjectPtrById(frm,curr2RateField);
		    currAmtFld = (FieldPtr)FrmGetObjectPtrById(frm,curr2AmountField);
		    restoreMainForm(frm);
		    clearMainForm(frm);
		}
	    } else if (event->data.frmLoad.formID == editListForm) {
		setTriggeredList( (ListPtr)FrmGetObjectPtrById( frm, ttceCategoryList),
				(ControlPtr)FrmGetObjectPtrById(frm, ttceCategoryTrigger),
				currentCategory);
		initChooser(  (ListPtr)FrmGetObjectPtrById( frm, itemsAvailable ),
			      currentCategory, true );
		}
            return TRUE;

	  case frmOpenEvent:
	    frm = FrmGetActiveForm();
	    if (event->data.frmOpen.formID == mainForm) {
		FrmSetFocusById( frm, ( GetAmountFromField(tipFld) > 0) ? amountField : tipRateField );
	    } else if (event->data.frmLoad.formID == editListForm) {
		newP = FALSE;
	    }
	    FrmDrawForm(frm);
            return TRUE;

	  case menuEvent:
	    if ( lastField ) updateMainPanel( lastField );
	    currentField = lastField = NULL;
	    MenuEventHandler(event);
            return TRUE;

	  case penDownEvent:
	    lastField = currentField;
	    currentField = NULL;
	    return FALSE;

	  case fldEnterEvent:
	    currentField = event->data.fldEnter.fieldID;
	    FrmHandleEvent(FrmGetActiveForm(), event); // default action
            return TRUE;
	    
	  //case keyDownEvent: // keyDownEvent, instead of penUpEvent, taken from currex!
	    // but keyDown is it sent when a button or a trigger are selected?
	    //if ( ( 0x30 <= event->date.keyDown.chr <= 0x39 ) )

	  case penUpEvent:
	    //if ( lastField != currentField ) updateMainPanel( lastField );
	    if (lastField && (event->screenY > 20)) updateMainPanel( lastField ); // Ugly
	      FrmHandleEvent(FrmGetActiveForm(), event); // default action
            return TRUE;
	    
	  case popSelectEvent:
	   { UInt categ=TAXCAT;
	    frm = FrmGetActiveForm();
	    switch ( event->data.popSelect.listID ) { // selection
	      case ttceCategoryList: 
		currentCategory = event->data.popSelect.selection;
		initChooser(  (ListPtr)FrmGetObjectPtrById( frm, itemsAvailable ),
			      currentCategory, false );
		break;
	      case currList:
	      case curr2List:
		computeCurrRate();
		computeValues();
		break;
	      case tipList: 
		categ = TIPCAT;
	      case taxList:
	      case tax2List:
		{ 
		  recordH = findRecordByRank( event->data.popSelect.selection, categ );
		  currentItem = MemHandleLock(recordH);
		  PutAmountInField( FrmGetObjectPtrById( frm, event->data.popSelect.listID + 1), currentItem->value );
		  if ( event->data.popSelect.listID == tax2List )
			CtlSetValue( taxOnAll, currentItem->opt );
		  MemHandleUnlock(recordH);
		  computeValues();
		  break ;
		}};
	    FrmHandleEvent(FrmGetActiveForm(), event); // default action
            return TRUE;
		}

	  case ctlSelectEvent:
	    switch (event->data.ctlSelect.controlID)
	    {
	      case calcBtn:
		computeValues();
                return TRUE;

	      case recBtn:
		storePrefs(storedPrefs);
                return TRUE;

	      case resBtn:
	        restorePrefs(storedPrefs);
		FrmSetFocusById( FrmGetActiveForm(), amountField);
		computeValues();
                return TRUE;

	      case tax2CBox:
		computeValues();
		return TRUE;
	      

	      case clearBtn:
		frm = FrmGetActiveForm();
	        clearMainForm(frm);
		FrmSetFocusById( frm, amountField);
		FrmDrawForm(frm);
                return TRUE;

	      case okAboutBtn:
		FrmReturnToForm(mainForm);
                return TRUE;
	
	      case delSelectButton:
		{
		ListPtr list = (ListPtr)FrmGetObjectPtrById(FrmGetActiveForm(), itemsAvailable);
		// Cannot destroy the fisrt one!
		if ( LstGetSelection(list) == 0 )  FrmAlert(delFirstItem);
		else {
		  recIndex = findRecIndexByRank( LstGetSelection(list), currentCategory );
		  //DBTODO: check if the record has been sync'ed once
		  // Retrieve the date of the last sync of the db
		  //if ( !gSaveBackUp ) DmRemoveRecord( db, recIndex );
		  //else {
			  DmDeleteRecord( db, recIndex );
			  DmMoveRecord( db, recIndex, DmNumRecords( db ));
		  //};
		  initChooser(  list, currentCategory, false );
		  toupdate = true;
		  LstDrawList(list);
                }
                }
                return TRUE;
		
	      case newSelectButton:
		newP = TRUE;
		//no break or return
	      case modSelectButton:
		if ( !newP ) {
			recIndex = findRecIndexByRank( LstGetSelection((ListPtr)FrmGetObjectPtrById(FrmGetActiveForm(), itemsAvailable)), currentCategory );
			recordH = DmQueryNextInCategory( db, &recIndex, currentCategory );
			currentItem = MemHandleLock(recordH);
		}
		FrmPopupForm( taxEditForm + currentCategory );//taxEditForm==5100
                return TRUE;
		
	      case okSelectButton:
		if ( toupdate ) restoreMainForm( FrmGetFormPtr(mainForm) );
		FrmReturnToForm(mainForm);
                return TRUE;
	    }

	  default:
	    FrmHandleEvent(FrmGetActiveForm(), event);
            return TRUE;
	}
	return FALSE;
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    EventType evt;
    Word error, prefsSize;
    FormPtr frm;
    Long intVal;
    Int type;
    Char thouChr;
    Int16 prefversion;

    if (cmd != sysAppLaunchCmdNormalLaunch) return FALSE;

    LocGetNumberSeparators((NumberFormatType)PrefGetPreference(prefNumberFormat),
			   &thouChr, &deciChr);
    deciSep[0] = deciNoDime[0] = deciChr;

    prefsSize = sizeof(TtcePrefsType);
    storedPrefs = MemPtrNew(prefsSize);
    getCurrentPrefs( storedPrefs, prefsSize, true );
//    prefs = MemPtrNew(prefsSize);
//    getCurrentPrefs( prefs, prefsSize, false );
//    restorePrefs(prefs);

    openDB();

    frm = FrmInitForm(mainForm);
    FrmGotoForm(mainForm);

    do {
	EvtGetEvent(&evt, evtWaitForever);
	if (SysHandleEvent(&evt)) continue;
	if (MenuHandleEvent(NULL, &evt, &error)) continue;
	
        if (evt.eType == frmLoadEvent)
        {
                Word Id = evt.data.frmLoad.formID;
		frm = FrmInitForm(Id) ;
		FrmSetActiveForm(frm) ;
		switch (Id) {
		  case mainForm: 
		  case editListForm:
		  case aboutForm:
		    FrmSetEventHandler(frm, (FormEventHandlerPtr)MainFormEventHandler) ;
		    break;
		  default: //  taxEditForm, tipEditForm, currEditForm:
		    FrmSetEventHandler(frm, (FormEventHandlerPtr)EditFormEventHandler) ;
		}
        } //else FrmHandleEvent(FrmGetActiveForm(), &evt);
	FrmDispatchEvent(&evt);
    } while ( evt.eType != appStopEvent );

    PrefSetAppPreferences(CREATOR, PREF_ID, PREF_VERSION, storedPrefs, prefsSize, true);
    MemPtrFree(storedPrefs);
//    storePrefs(prefs);
//    PrefSetAppPreferences(CREATOR, PREF_ID, PREF_VERSION, prefs, prefsSize, false);
//    MemPtrFree(prefs);

    DmCloseDatabase(db);

    // Unset the handlers for all the fields...
    FldSetTextHandle(amountFld,NULL);
    FldSetTextHandle(tipFld,NULL);
    FldSetTextHandle(taxFld,NULL);
    FldSetTextHandle(tax2Fld,NULL);
    FldSetTextHandle(totalFld,NULL);
    FldSetTextHandle(currFld,NULL);
    FrmCloseAllForms();

    return FALSE;
}
